from typing import Optional, Any
from temporalio import workflow
from temporalio.common import RetryPolicy
from temporalio.exceptions import ActivityError
from dataclasses import dataclass
from datetime import timedelta
from collections import deque, defaultdict
from admyral.utils.collections import is_not_empty
import asyncio
from pydantic import BaseModel, JsonValue
from copy import deepcopy

# Import activity, passing it through the sandbox without reloading the module
with workflow.unsafe.imports_passed_through():
    from admyral.logger import get_logger
    from admyral.models import (
        ActionNode,
        IfNode,
        WorkflowDAG,
        Workflow,
    )
    from admyral.action_registry import ActionRegistry
    from admyral.workers.references import evaluate_references
    from admyral.workers.if_condition_executor import ConditionReferenceResolution


logger = get_logger(__name__)


START_TO_CLOSE_TIMEOUT = timedelta(seconds=60 * 15)  # 15 min
ACTION_RETRY_POLICY = RetryPolicy(
    maximum_attempts=3,
    non_retryable_error_types=["NonRetryableActionError"],
)


# NOTE: params as objects are strongly encouraged
# see https://docs.temporal.io/develop/python/core-application#workflow-parameters
@dataclass
class WorkflowParams:
    workflow: Workflow
    source_name: str
    payload: dict[str, Any]
    trigger_default_args: dict[str, Any]


class JobQueueEntry(BaseModel):
    action_id: str
    prev_step_id: Optional[str] = None


# Workflow code requirements:
# - no threading
# - no randomness
# - no external calls to processes
# - no network I/O
# - no global state mutation
# - no system date or time
#
# run methods only accept positional parameters


async def _execute_activity(action_type: str, args: list[JsonValue]) -> JsonValue:
    try:
        return await workflow.execute_activity(
            action_type,
            args=args,
            start_to_close_timeout=START_TO_CLOSE_TIMEOUT,
            retry_policy=ACTION_RETRY_POLICY,
        )
    except ActivityError as e:
        logger.error(f"Error executing activity of type {action_type}: {e.message}")
        raise e


@workflow.defn(name="WorkflowExecutor")
class WorkflowExecutor:
    @workflow.run
    async def run(self, params: WorkflowParams) -> None:
        workflow_dag = params.workflow.workflow_dag
        in_deg = workflow_dag.get_in_deg()

        # initialize workflow run
        payload = self._inject_default_args(params.payload, params.trigger_default_args)

        workflow_run_id = await _execute_activity(
            "init_workflow_run",
            args=[params.workflow.workflow_id, params.source_name, payload],
        )

        logger.info(
            f'Triggering workflow "{params.workflow.workflow_id}" from source "{params.source_name}" with run ID "{workflow_run_id}".'
        )

        # setup states for workflow execution
        execution_state = {"payload": payload}

        eliminated_nodes = set()
        resolved_dependencies = defaultdict(int)
        number_of_executed_actions = 0

        job_queue = asyncio.Queue()
        exception_queue = asyncio.Queue()

        async def task(action_id: str, prev_step_id: str):
            nonlocal number_of_executed_actions
            nonlocal eliminated_nodes
            nonlocal resolved_dependencies

            try:
                node = params.workflow.workflow_dag.dag[action_id]

                # TODO: strong type?
                ctx_dict = {
                    "workflow_id": params.workflow.workflow_id,
                    "run_id": workflow_run_id,
                    "action_type": node.type,
                    "prev_step_id": prev_step_id,
                }

                logger.info(f"Executing action: {action_id}")

                number_of_executed_actions += 1

                # execution
                if isinstance(node, IfNode):
                    step_id, newly_eliminated_nodes = await self._execute_if_condition(
                        node,
                        execution_state,
                        params.workflow.workflow_dag,
                        in_deg,
                        ctx_dict,
                    )
                    eliminated_nodes |= newly_eliminated_nodes
                    candidates = node.true_children + node.false_children
                elif isinstance(node, ActionNode):
                    if node.id != "start":
                        if node.type == "wait":
                            await asyncio.sleep(node.args.get("seconds", 0))

                        step_id, execution_result = await self._execute_action_node(
                            node, execution_state, ctx_dict
                        )
                        if node.result_name is not None:
                            execution_state[node.result_name] = execution_result
                    else:
                        step_id = None
                    candidates = node.children
                else:
                    raise RuntimeError(f"Invalid node type: {type(node)}")

                # schedule next actions
                for child_id in candidates:
                    # mark the current node as resolved for each child
                    resolved_dependencies[child_id] += 1
                    # we only schedule a child if all its dependencies (i.e., its parents) are resolved
                    if (
                        child_id not in eliminated_nodes
                        and resolved_dependencies[child_id] == in_deg[child_id]
                    ):
                        await job_queue.put(
                            JobQueueEntry(action_id=child_id, prev_step_id=step_id)
                        )

                # if now all tasks were scheduled, notify main execution loop that we are now done
                if number_of_executed_actions + len(eliminated_nodes) == len(
                    params.workflow.workflow_dag.dag
                ):
                    await job_queue.put(None)

            except Exception as e:
                logger.error(f"Error executing task: {action_id}. Exception: {e}")
                await job_queue.put(None)
                await exception_queue.put(e)

            job_queue.task_done()

        # job trigger loop
        await job_queue.put(JobQueueEntry(action_id="start", prev_step_id=None))

        exception = None
        async with asyncio.TaskGroup() as tg:
            while True:
                job = await job_queue.get()

                # Check for exceptions
                try:
                    exception = exception_queue.get_nowait()
                    # an exception occurred! we stop the scheduling loop
                    # but we let the currently running tasks complete
                    break
                except asyncio.QueueEmpty:
                    pass

                if job is None:
                    break

                # launch new task
                logger.info(
                    f"Scheduling action: {job.action_id}"
                )  # TODO: better log message
                tg.create_task(task(job.action_id, job.prev_step_id))

        if exception:
            raise exception

        await _execute_activity("mark_workflow_as_completed", args=[workflow_run_id])

        logger.info(
            f'Workflow execution of workflow "{params.workflow.workflow_id}" with run ID "{workflow_run_id}" completed successfully.'
        )

    def _inject_default_args(
        self,
        payload: dict[str, JsonValue],
        trigger_default_args: dict[str, JsonValue],
    ) -> dict[str, JsonValue]:
        # We have the following value priority:
        # 1. Value in payload
        # 2. Default value in trigger (already handled before, i.e., during scheduling or on API server-side for webhooks)

        # handle trigger default args
        for arg_name, arg_value in trigger_default_args.items():
            if arg_name not in payload:
                payload[arg_name] = arg_value

        return payload

    async def _execute_action_node(
        self, node: ActionNode, execution_state: dict, ctx_dict: dict[str, Any]
    ) -> tuple[str, Any]:
        # evaluate the references of the action arguments
        action_args = {}
        for key, value in node.args.items():
            key = str(
                evaluate_references(key, execution_state)
            )  # keys of JSON objects always must be strings!
            value = evaluate_references(value, execution_state)
            action_args[key] = value

        action_type = node.type

        if action_type not in ActionRegistry.get_action_types():
            # Custom Python action
            return await _execute_activity(
                "execute_python_action",
                args=[
                    ctx_dict,
                    node.secrets_mapping,
                    {"action_type": action_type, "action_args": action_args},
                ],
            )
        return await _execute_activity(
            action_type,
            args=[ctx_dict, node.secrets_mapping, action_args],
        )

    async def _execute_if_condition(
        self,
        dag_node: IfNode,
        execution_state: dict,
        workflow_dag: WorkflowDAG,
        in_deg: dict[str, int],
        ctx_dict: dict[str, Any],
    ) -> tuple[str, set[str]]:
        # 1) evaluate if-condition

        # if we perform model_dump directly on the condition, we get a circular reference error for
        # some reason. Doing a model dump on an if-node works fine, though.
        # TODO: we should probably try to fix this in the future because ugly
        dag_node_copy = deepcopy(dag_node)
        dag_node_copy.condition = ConditionReferenceResolution(
            execution_state
        ).resolve_references(dag_node_copy.condition)
        dag_node_copy_json = dag_node_copy.model_dump()
        condition_json = dag_node_copy_json["condition"]

        step_id, condition_result = await _execute_activity(
            "if_condition",
            args=[ctx_dict, {}, {"condition_expr": condition_json}],
        )

        # 2) path elimination: remove the untaken path from the execution order
        eliminated_nodes = set()
        if condition_result:
            eliminated_nodes |= path_elimination(
                dag_node.false_children, workflow_dag, in_deg
            )
        else:
            eliminated_nodes |= path_elimination(
                dag_node.true_children, workflow_dag, in_deg
            )

        return step_id, eliminated_nodes


def path_elimination(
    node_ids: list[str], workflow_dag: WorkflowDAG, in_deg: dict[str, int]
) -> set[str]:
    """
    Path Elimination for if-conditions.

    Consider the following DAG:

              ...
              ||
            if cond
          //        \\
    action1          action2
          \\        //
            action3
              ||
              ...

    In our execution model, we execute an action if all its dependencies are resolved, i.e.,
    incoming degree is 0. Therefore, we must reduce the incoming degree of action3 in order to
    be able to execute it. This is done by eliminating the path that is not taken by the if-condition.
    """
    eliminated_nodes = set()

    # BFS traversal for elimination - don't expand nodes with in_deg > 0
    queue = deque(node_ids)
    while is_not_empty(queue):
        current_node_id = queue.popleft()
        eliminated_nodes.add(current_node_id)

        # reduce in_deg of children
        node = workflow_dag.dag[current_node_id]
        if isinstance(node, IfNode):
            for child_id in node.true_children:
                in_deg[child_id] -= 1
                # remove child if it does not have a dependency anymore (i.e., in_deg == 0)
                if in_deg[child_id] == 0:
                    queue.append(child_id)
            for child_id in node.false_children:
                in_deg[child_id] -= 1
                # remove child if it does not have a dependency anymore (i.e., in_deg == 0)
                if in_deg[child_id] == 0:
                    queue.append(child_id)
        else:
            for child_id in node.children:
                in_deg[child_id] -= 1
                # remove child if it does not have a dependency anymore (i.e., in_deg == 0)
                if in_deg[child_id] == 0:
                    queue.append(child_id)

    return eliminated_nodes



from admyral.workers.worker import run_worker
from admyral.config.config import GlobalConfig


async def launch_worker(args: dict) -> None:
    # if args.get("docker"):
    #     target = "host.docker.internal:7233"
    # else:
    #     target = "127.0.0.1:7233"
    worker_name = "admyral-worker"
    await run_worker(worker_name, GlobalConfig().temporal_host, "workflow-queue")



from typing import Any


def is_empty(collection: dict | list | Any) -> bool:
    return len(collection) == 0


def is_not_empty(collection: dict | list | Any) -> bool:
    return len(collection) > 0



import click
import os
import subprocess

from admyral.cli.cli import cli
from admyral.utils.docker_utils import (
    is_docker_running,
    get_docker_compose_cmd,
    list_running_docker_containers,
)
from admyral.config.config import get_local_postgres_volume
from admyral import __version__


WELCOME_MESSAGE = """

 _       __     __                             __      
| |     / /__  / /________  ____ ___  ___     / /_____ 
| | /| / / _ \\/ / ___/ __ \\/ __ `__ \\/ _ \\   / __/ __ \\
| |/ |/ /  __/ / /__/ /_/ / / / / / /  __/  / /_/ /_/ /
|__/|__/\\___/_/\\___/\\____/_/ /_/ /_/\\___/  _\\__/\\____/ 
   /   | ____/ /___ ___  __  ___________ _/ /          
  / /| |/ __  / __ `__ \\/ / / / ___/ __ `/ /           
 / ___ / /_/ / / / / / / /_/ / /  / /_/ / /            
/_/  |_\\__,_/_/ /_/ /_/\\__, /_/   \\__,_/_/             
                      /____/                           

"""


def _get_docker_compose_dir_path() -> str:
    if os.path.exists(os.path.join(os.path.dirname(__file__), "docker_compose")):
        docker_compose_dir_path = os.path.join(
            os.path.dirname(__file__), "docker_compose"
        )
    else:
        docker_compose_dir_path = os.path.join(
            os.path.dirname(__file__), "..", "..", "deploy", "docker-compose"
        )
    assert os.path.exists(docker_compose_dir_path)
    return docker_compose_dir_path


@cli.command("up", help="Start Admyral locally.")
def up() -> None:
    """
    Launches Admyral services locally.
    """
    click.echo(WELCOME_MESSAGE)

    click.echo(f"Admyral Version: {__version__}\n")

    if not is_docker_running():
        click.echo(
            "Docker daemon is not running. Please make sure that Docker is installed and running."
        )
        return

    # check if container is already running
    running_containers = list_running_docker_containers()
    all_services = {
        "admyral-web",
        "admyral-worker",
        "admyral-api",
        "temporal-admin-tools",
        "temporal-ui",
        "temporal",
        "postgresql",
        "temporal-elasticsearch",
    }
    if len(set(running_containers) & all_services) == len(all_services):
        click.echo("Admyral is already running.")
        click.echo(
            "You can access the Admyral UI at http://localhost:3000 or use the Admyral CLI.\n"
        )
        return

    # figure out the path of the docker-compose directory
    docker_compose_dir_path = _get_docker_compose_dir_path()

    command = get_docker_compose_cmd()
    command.append("up")
    command.append("-d")

    env = os.environ.copy()
    if openai_api_key := os.environ.get("OPENAI_API_KEY"):
        env["OPENAI_API_KEY"] = openai_api_key
    else:
        click.echo(
            'Warning: OPENAI_API_KEY environment variable is not set. The "ai_action" action will not work!'
        )
    if (
        os.environ.get("RESEND_API_KEY") is None
        or os.environ.get("RESEND_EMAIL") is None
    ):
        click.echo(
            'Warning: RESEND_API_KEY or RESEND_EMAIL environment variables are not set. The "send_email" action will not work!'
        )
    else:
        env["RESEND_API_KEY"] = os.environ.get("RESEND_API_KEY")
        env["RESEND_EMAIL"] = os.environ.get("RESEND_EMAIL")

    # Set persistance path
    env["POSTGRES_VOLUME_PATH"] = get_local_postgres_volume()

    env["ADMYRAL_VERSION"] = __version__

    click.echo("\nStarting Admyral...\n")

    try:
        subprocess.run(command, check=True, cwd=docker_compose_dir_path, env=env)
    except subprocess.CalledProcessError as e:
        click.echo(f"Command failed with error: {e}")
        return

    click.echo("\nAdmyral is up and running.")
    click.echo(
        "You can access the Admyral UI at http://localhost:3000 or use the Admyral CLI.\n"
    )


@cli.command("down", help="Stop Admyral locally.")
def down() -> None:
    """
    Tears down Admyral services locally.
    """
    if not is_docker_running():
        click.echo(
            "Docker daemon is not running. Please make sure that Docker is installed and running."
        )
        return

    docker_compose_dir_path = _get_docker_compose_dir_path()
    command = get_docker_compose_cmd()
    command.append("down")

    # Set persistance path
    env = os.environ.copy()
    env["POSTGRES_VOLUME_PATH"] = get_local_postgres_volume()
    env["ADMYRAL_VERSION"] = __version__

    click.echo("\nShutting Admyral down...\n")

    try:
        subprocess.run(command, check=True, cwd=docker_compose_dir_path, env=env)
    except subprocess.CalledProcessError as e:
        click.echo(f"Command failed with error: {e}")
        return

    click.echo("\nAdmyral is shut down.\n")


# @cli.command("show", help="Show information about Admyral.")
# def show() -> None:
#     """
#     Show information about Admyral.
#     """
#     show_status()


# @cli.command("logs", help="Display logs for Admyral.")
# @click.option("--follow", "-f", is_flag=True, help="Follow the logs.")
# @click.option(
#     "--tail",
#     "-t",
#     type=click.INT,
#     help="Number of lines to display from the end of the logs.",
# )
# def logs(follow: bool, tail: int) -> None:
#     """
#     Display logs for Admyral.

#     Args:
#         follow: Follow the logs.
#         tail: Number of lines to display from the end of the logs.
#     """
#     for log in show_logs(follow, tail):
#         click.echo(log)



from typing import Annotated
from httpx import Client

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue


def get_abnormal_security_client(api_key: str) -> Client:
    return Client(
        base_url="https://api.abnormalplatform.com/v1",
        headers={
            "Authorization": f"Bearer {api_key}",
            "Accept": "application/json",
        },
    )


# TODO: normalize cases (OCSF?)
@action(
    display_name="List Abnormal Cases",
    display_namespace="Abnormal Security",
    description="Get a list of Abnormal cases identified.",
    secrets_placeholders=["ABNORMAL_SECURITY_SECRET"],
)
def list_abnormal_security_cases(
    start_time: Annotated[
        str,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = "1970-01-01T00:00:00Z",
    end_time: Annotated[
        str,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = "2100-01-01T00:00:00Z",
    limit: Annotated[
        int,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of cases to list.",
        ),
    ] = 1000,
) -> list[dict[str, JsonValue]]:
    # https://app.swaggerhub.com/apis/abnormal-security/abx/1.4.3#/Cases/get_cases

    secret = ctx.get().secrets.get("ABNORMAL_SECURITY_SECRET")
    api_key = secret["api_key"]

    filter = f"createdTime gte {start_time} lte {end_time}"

    with get_abnormal_security_client(api_key) as client:
        page_number = 1
        cases = []

        while len(cases) < limit:
            response = client.get(
                "/cases",
                params={
                    "pageSize": 100,
                    "pageNumber": page_number,
                    "filter": filter,
                },
            )
            response.raise_for_status()
            result = response.json()

            cases.extend(result.get("cases", []))

            if "nextPageNumber" in result:
                page_number = result["nextPageNumber"]
            else:
                break

        return cases[:limit]



import time
import aiofiles
import aiofiles.os
import asyncio
from typing import Any
import tempfile
import json
import os
import sys
import itertools

from admyral.utils.aio import makedirs, dirname
from admyral.config.config import (
    ADMYRAL_PIP_CACHE_DIRECTORY,
    ADMYRAL_PIP_LOCK_CACHE_DIRECTORY,
    ADMYRAL_DISABLE_NSJAIL,
    ADMYRAL_PIP_LOCKFILE_CACHE_TTL_IN_SECONDS,
    ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE,
)
from admyral.workers.shared_worker_state import SharedWorkerState
from admyral.logger import get_logger
from admyral.models import PythonAction, PipLockfile
from admyral.db.store_interface import StoreInterface
from admyral.utils.hash import calculate_sha256
from admyral.utils.time import utc_now_timestamp_seconds
from admyral.utils.aio import path_exists, getcwd
from admyral.utils.subprocess import run_subprocess_with_log_flushing
from admyral.context import ctx

logger = get_logger(__name__)


MOUNT_TEMPLATE = """mount {{
    src: \"{REQUIREMENT_PATH}\"
    dst: \"{REQUIREMENT_PATH}\"
    is_bind: true
}}"""


ADMYRAL_PYTHON_PATH = "/usr/local/bin/python"


async def get_nsjail_dir() -> str:
    file_dir = await dirname(__file__)
    return os.path.join(file_dir, "nsjail")


async def python_action_worker_setup() -> None:
    if not ADMYRAL_DISABLE_NSJAIL:
        await makedirs(ADMYRAL_PIP_CACHE_DIRECTORY)
        await makedirs(ADMYRAL_PIP_LOCK_CACHE_DIRECTORY)


async def execute_python_action(action_type: str, action_args: dict[str, Any]) -> Any:
    logger.info(f"Executing Python action with type '{action_type}'")

    job_start = time.monotonic_ns()

    store = SharedWorkerState.get_store()

    # Load the action from the store
    python_action = await store.get_action(action_type)
    if not python_action:
        raise RuntimeError(
            f"Action with type '{action_type}' not found. Did you push your action?"
        )

    with tempfile.TemporaryDirectory() as job_dir:
        logger.info(f"Job directory: {job_dir}")

        # write Python action and requirements file
        await _prepare_python_action(job_dir, python_action, action_args)

        await _run_python_action(job_dir, python_action, store)

        # Read output from file
        async with aiofiles.open(os.path.join(job_dir, "output.json"), "r") as f:
            output = await f.read()
            if output == "":
                result = None
            else:
                result = json.loads(output)

    job_end = time.monotonic_ns()
    logger.info(
        f"Python action with type '{action_type}' executed in {(job_end - job_start) / 1_000_000} ms"
    )

    return result


async def _load_secrets(python_action: PythonAction) -> dict[str, str]:
    loaded_secrets = {}
    for secret_placeholder in python_action.secrets_placeholders:
        secret = await ctx.get().secrets.aget(secret_placeholder)
        loaded_secrets[secret_placeholder] = json.dumps(secret)
    return loaded_secrets


def _build_python_action_module(python_action: PythonAction) -> str:
    return f"{python_action.import_statements}\n\n{python_action.code}"


async def _prepare_python_action(
    job_dir: str, python_action: PythonAction, action_args: dict[str, Any]
) -> None:
    # prepare input and output files
    async with aiofiles.open(os.path.join(job_dir, "input.json"), "w") as f:
        await f.write(json.dumps(action_args))

    async with aiofiles.open(os.path.join(job_dir, "output.json"), "w") as f:
        await f.write("")

    # Write requirements file
    requirements_list = python_action.requirements
    if "admyral" not in requirements_list:
        requirements_list.append("admyral")

    if ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE:
        # remove admyral from requirements if we are using the local package
        requirements_list = [req for req in requirements_list if req != "admyral"]

    # filter out packages from the standard library
    requirements_list = [
        req
        for req in requirements_list
        if req.split("==")[0] not in sys.stdlib_module_names
    ]

    async with aiofiles.open(os.path.join(job_dir, "requirements.in"), "w") as f:
        await f.write("\n".join(requirements_list))

    # Write Python files
    async with aiofiles.open(os.path.join(job_dir, "action.py"), "w") as f:
        action_module = _build_python_action_module(python_action)
        await f.write(action_module)

    nsjail_dir = await get_nsjail_dir()
    async with aiofiles.open(
        os.path.join(nsjail_dir, "template.python_action_executor.py.txt")
    ) as f:
        python_action_executor_template = await f.read()

    python_action_executor = python_action_executor_template.format(
        ACTION_PATH="action", ACTION_TYPE=python_action.action_type, JOB_DIR=job_dir
    )

    async with aiofiles.open(
        os.path.join(job_dir, "python_action_executor.py"), "w"
    ) as f:
        await f.write(python_action_executor)


async def _run_python_action(
    job_dir: str,
    python_action: PythonAction,
    store: StoreInterface,
) -> None:
    if ADMYRAL_DISABLE_NSJAIL:
        await _run_python_action_without_nsjail(job_dir, python_action)
    else:
        await _run_python_action_with_nsjail(job_dir, python_action, store)


async def _run_python_action_without_nsjail(
    job_dir: str,
    python_action: PythonAction,
) -> None:
    python_path = sys.executable

    logger.info(f"Python Path: {python_path}")

    # load secrets as environment for subprocess
    env = await _load_secrets(python_action)

    cmd = [
        python_path,
        "-u",  # unbuffered binary stdout and stderr
        os.path.join(job_dir, "python_action_executor.py"),
    ]
    exit_code = await run_subprocess_with_log_flushing(
        cmd, lambda logs: ctx.get().append_logs_async(logs), env=env
    )

    _handle_exit_code(exit_code, cmd, python_action.action_type)


async def _run_python_action_with_nsjail(
    job_dir: str,
    python_action: PythonAction,
    store: StoreInterface,
) -> None:
    lockfile = await _pip_compile(job_dir, python_action, store)
    requirements_paths = await _pip_install_requirements(
        python_action.action_type, lockfile, job_dir
    )
    await _jailed_python_execution(requirements_paths, job_dir, python_action)


async def _pip_compile(
    job_dir: str, python_action: PythonAction, store: StoreInterface
) -> list[str]:
    """Generate lockfile using pip-compile"""
    hash_id = calculate_sha256(";".join(python_action.requirements))
    pip_lockfile = await store.get_cached_pip_lockfile(hash_id)
    if pip_lockfile:
        # we still have a valid lockfile in our cache
        logger.info("Skipping pip-compile. Found cached lockfile.")
        return pip_lockfile.lockfile.split("\n")

    pip_compile_start = time.monotonic_ns()

    # we need to generate a new lockfile
    logger.info("Generating lockfile using pip-compile...")

    cmd = [
        "pip-compile",
        "-q",
        "--no-header",
        os.path.join(job_dir, "requirements.in"),
        "--resolver=backtracking",
        "--strip-extras",
    ]
    exit_code = await run_subprocess_with_log_flushing(
        cmd, lambda logs: ctx.get().append_logs_async(logs)
    )

    _handle_exit_code(exit_code, cmd, python_action.action_type)

    lockfile_path = os.path.join(job_dir, "requirements.txt")
    async with aiofiles.open(lockfile_path) as f:
        lockfile = await f.readlines()

    lockfile = list(
        filter(
            lambda req: not req.startswith("#"),
            map(lambda req: req.strip().replace("\n", ""), lockfile),
        )
    )

    pip_compiled_end = time.monotonic_ns()
    logger.info(
        f"Pip Compile Time: {(pip_compiled_end - pip_compile_start) / 1_000_000} ms"
    )

    # store lock file in cache
    await store.cache_pip_lockfile(
        PipLockfile(
            hash=hash_id,
            lockfile="\n".join(lockfile),
            expiration_time=utc_now_timestamp_seconds()
            + ADMYRAL_PIP_LOCKFILE_CACHE_TTL_IN_SECONDS,
        )
    )

    return lockfile


async def _pip_install_requirements(
    action_type: str, lockfile: list[str], job_dir: str
) -> list[str]:
    pip_install_start = time.monotonic_ns()

    installed_requirements_paths = []
    if ADMYRAL_USE_LOCAL_ADMYRAL_PIP_PACKAGE:
        # add the app directory to the installed requirements, such that
        # the action can access the local package
        logger.info("Using local Admyral package")
        cwd = await getcwd()
        installed_requirements_paths.append(cwd)
    else:
        logger.info("Using PyPi Admyral package")

    missing_requirements = []
    for requirement in lockfile:
        requirement_cache_path = os.path.join(ADMYRAL_PIP_CACHE_DIRECTORY, requirement)

        if await path_exists(requirement_cache_path):
            installed_requirements_paths.append(requirement_cache_path)
        else:
            missing_requirements.append((requirement, requirement_cache_path))

    cwd = await getcwd()
    install_requirement_script = os.path.join(
        cwd, "admyral", "workers", "nsjail", "install_requirement.sh"
    )

    installed_requirements_paths += await asyncio.gather(
        *[
            _pip_install_requirement(
                action_type,
                requirement,
                requirement_cache_path,
                job_dir,
                install_requirement_script,
            )
            for requirement, requirement_cache_path in missing_requirements
        ]
    )

    pip_install_end = time.monotonic_ns()
    logger.info(
        f"Pip Install Time: {(pip_install_end - pip_install_start) / 1_000_000} ms"
    )

    return installed_requirements_paths


async def _pip_install_requirement(
    action_type: str,
    requirement: str,
    requirement_cache_path: str,
    job_dir: str,
    install_requirement_script: str,
) -> str:
    lock_name = requirement.replace("==", "_").replace(".", "_").replace("-", "_")
    req_lock_file = os.path.join(
        ADMYRAL_PIP_LOCK_CACHE_DIRECTORY, f"pip-{lock_name}.lock"
    )

    # generate nsjail config file
    async with aiofiles.open(
        os.path.join("admyral", "workers", "nsjail", "template.pip_install.cfg")
    ) as f:
        nsjail_config_template = await f.read()

    nsjail_config = nsjail_config_template.format(
        FLOCK=req_lock_file,
        REQUIREMENT=requirement,
        TARGET=requirement_cache_path,
        INSTALL_REQUIREMENT_SCRIPT_PATH=install_requirement_script,
    )
    nsjail_config_path = os.path.join(job_dir, f"nsjail_pip_install_{requirement}.cfg")
    async with aiofiles.open(nsjail_config_path, "w") as f:
        await f.write(nsjail_config)

    cmd = [
        "nsjail",
        "--config",
        nsjail_config_path,
    ]
    exit_code = await run_subprocess_with_log_flushing(
        cmd, lambda logs: ctx.get().append_logs_async(logs)
    )

    _handle_exit_code(exit_code, cmd, action_type)

    return requirement_cache_path


async def _jailed_python_execution(
    requirements_paths: list[str],
    job_dir: str,
    python_action: PythonAction,
) -> None:
    execution_start = time.monotonic_ns()

    logger.info("Executing Python action with nsjail")
    x = ":".join(requirements_paths)
    logger.info(f"Requirements paths: {x}")

    requirements_mounts = "\n".join(
        [
            MOUNT_TEMPLATE.format(REQUIREMENT_PATH=requirement_path)
            for requirement_path in requirements_paths
        ]
    )

    async with aiofiles.open(
        os.path.join("admyral", "workers", "nsjail", "template.python_action.cfg")
    ) as f:
        nsjail_config_template = await f.read()

    nsjail_config = nsjail_config_template.format(
        EXECUTOR_PATH=os.path.join(job_dir, "python_action_executor.py"),
        ACTION_PATH=os.path.join(job_dir, "action.py"),
        PYTHON_DEPENDENCIES=":".join(requirements_paths),
        PYTHON_DEPENDENCIES_MOUNT=requirements_mounts,
        JOB_DIR=job_dir,
    )

    nsjail_config_path = os.path.join(job_dir, "action.cfg")
    async with aiofiles.open(nsjail_config_path, "w") as f:
        await f.write(nsjail_config)

    # load secrets and inject as environment variables
    env = await _load_secrets(python_action)

    cmd = [
        "nsjail",
        "--config",
        nsjail_config_path,
        "--",
        ADMYRAL_PYTHON_PATH,
        "-u",  # unbuffered binary stdout and stderr
        "python_action_executor.py",
    ]
    # inject secrets by appending them as command line arguments
    cmd_with_secrets = cmd + list(
        itertools.chain.from_iterable(
            [["-s", f"'{len(key)}|{key}{value}'"] for key, value in env.items()]
        )
    )
    exit_code = await run_subprocess_with_log_flushing(
        cmd_with_secrets,
        lambda logs: ctx.get().append_logs_async(logs),
    )

    _handle_exit_code(exit_code, cmd, python_action.action_type)

    execution_end = time.monotonic_ns()
    logger.info(
        f"Python action executed in {(execution_end - execution_start) / 1_000_000} ms"
    )


def _handle_exit_code(exit_code: int, cmd: list[str], action_type: str):
    if exit_code != 0:
        cmd_str = " ".join(cmd)
        logger.error(
            f'Failed execute command "{cmd_str}". Exited with code {exit_code}.'
        )
        raise RuntimeError(
            f"Failed to execute action: {action_type}. See logs for more details."
        )



from typing import Annotated
from httpx import Client

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue


def get_grey_noise_client(api_key: str) -> Client:
    return Client(
        base_url="https://api.greynoise.io/v3",
        headers={
            "x-apikey": api_key,
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Analyze IP Address",
    display_namespace="GreyNoise",
    description="Analyze an IP address using GreyNoise",
    secrets_placeholders=["GREY_NOISE_SECRET"],
)
def grey_noise_ip_lookup(
    ip_address: Annotated[
        str,
        ArgumentMetadata(
            display_name="IP Address", description="The IP address to analyze"
        ),
    ],
) -> JsonValue:
    secret = ctx.get().secrets.get("GREY_NOISE_SECRET")
    api_key = secret["api_key"]
    with get_grey_noise_client(api_key) as client:
        response = client.get(f"/community/{ip_address}")
        response.raise_for_status()
        return response.json()



from admyral.models import (
    Workflow,
    EditorWorkflowGraph,
    EditorWorkflowActionNode,
    EditorWorkflowIfNode,
    EditorWorkflowStartNode,
    EditorWorkflowEdge,
    EditorWorkflowEdgeType,
    IfNode,
    ActionNode,
    WorkflowStart,
    WorkflowDAG,
    WorkflowTriggerType,
    EditorWebhookTrigger,
    EditorScheduleTrigger,
    WorkflowWebhookTrigger,
    WorkflowScheduleTrigger,
    WorkflowDefaultArgument,
    EditorScheduleType,
)
from admyral.compiler.condition_compiler import compile_condition_str
from admyral.editor.json_with_references_serde import (
    serialize_json_with_reference,
    deserialize_json_with_reference,
)
from admyral.utils.collections import is_not_empty


def workflow_to_editor_workflow_graph(
    workflow: Workflow, webhook_id: str | None, webhook_secret: str | None
) -> EditorWorkflowGraph:
    """
    Transform Workflow to editor workflow graph.

    Args:
        workflow: Workflow to transform.

    Returns:
        editor workflow graph.
    """
    webhook_trigger = None
    schedule_triggers = []
    for trigger in workflow.workflow_dag.start.triggers:
        match trigger.type:
            case WorkflowTriggerType.WEBHOOK:
                if webhook_trigger:
                    raise ValueError("Multiple webhook triggers found.")

                webhook_trigger = EditorWebhookTrigger(
                    webhook_id=webhook_id,
                    webhook_secret=webhook_secret,
                    default_args=[
                        (k, serialize_json_with_reference(v))
                        for k, v in trigger.default_args_dict.items()
                    ],
                )

            case WorkflowTriggerType.SCHEDULE:
                schedule_triggers.append(_build_editor_schedule_trigger(trigger))

            case _:
                raise ValueError(f"Unhandled trigger type: {trigger.type}")

    nodes = []
    edges = []

    for node_id, node in workflow.workflow_dag.dag.items():
        if isinstance(node, IfNode):
            # IF CONDITION NODES
            nodes.append(EditorWorkflowIfNode(id=node_id, condition=node.condition_str))
            for true_child in node.true_children:
                edges.append(
                    EditorWorkflowEdge(
                        source=node_id,
                        target=true_child,
                        type=EditorWorkflowEdgeType.TRUE,
                    )
                )
            for false_child in node.false_children:
                edges.append(
                    EditorWorkflowEdge(
                        source=node_id,
                        target=false_child,
                        type=EditorWorkflowEdgeType.FALSE,
                    )
                )
        elif node.id == "start":
            # START NODES
            nodes.append(
                EditorWorkflowStartNode(
                    id=node_id, webhook=webhook_trigger, schedules=schedule_triggers
                )
            )
            for child in node.children:
                edges.append(
                    EditorWorkflowEdge(
                        source=node_id,
                        target=child,
                        type=EditorWorkflowEdgeType.DEFAULT,
                    )
                )
        else:
            # ACTION NODES
            nodes.append(
                EditorWorkflowActionNode(
                    id=node_id,
                    action_type=node.type,
                    result_name=node.result_name,
                    secrets_mapping=node.secrets_mapping,
                    args={
                        k: serialize_json_with_reference(v)
                        for k, v in node.args.items()
                    },
                )
            )
            for child in node.children:
                edges.append(
                    EditorWorkflowEdge(
                        source=node_id,
                        target=child,
                        type=EditorWorkflowEdgeType.DEFAULT,
                    )
                )

    return EditorWorkflowGraph(
        workflow_id=workflow.workflow_id,
        workflow_name=workflow.workflow_name,
        description=workflow.workflow_dag.description,
        is_active=workflow.is_active,
        nodes=nodes,
        edges=edges,
    )


def editor_workflow_graph_to_workflow(
    editor_workflow_graph: EditorWorkflowGraph,
) -> Workflow:
    """
    Transform editor workflow graph to Workflow.

    Args:
        editor_workflow_graph: Editor workflow graph to transform.

    Returns:
        Workflow.
    """
    workflow_dag = {}
    triggers = []

    for node in editor_workflow_graph.nodes:
        if isinstance(node, EditorWorkflowStartNode):
            if node.id in workflow_dag:
                raise ValueError("Multiple start nodes found.")
            workflow_dag[node.id] = ActionNode.build_start_node()

            if node.webhook:
                triggers.append(
                    WorkflowWebhookTrigger(
                        default_args=[
                            WorkflowDefaultArgument(
                                name=k, value=deserialize_json_with_reference(v)
                            )
                            for k, v in node.webhook.default_args
                            # If value is an empty string, then there is no value
                            # and we can skip it.
                            if is_not_empty(v)
                        ]
                    )
                )

            for schedule in node.schedules:
                triggers.append(_build_workflow_schedule_trigger(schedule))

            continue

        if isinstance(node, EditorWorkflowActionNode):
            workflow_dag[node.id] = ActionNode(
                id=node.id,
                type=node.action_type,
                result_name=node.result_name,
                secrets_mapping=node.secrets_mapping,
                args={
                    k: deserialize_json_with_reference(v)
                    for k, v in node.args.items()
                    # If value is an empty string, then there is no value
                    # and we can skip it.
                    if is_not_empty(v)
                },
            )
            continue

        if isinstance(node, EditorWorkflowIfNode):
            workflow_dag[node.id] = IfNode(
                id=node.id,
                condition=compile_condition_str(node.condition),
                condition_str=node.condition,
            )
            continue

    for edge in editor_workflow_graph.edges:
        match edge.type:
            case EditorWorkflowEdgeType.DEFAULT:
                workflow_dag[edge.source].add_edge(edge.target)
            case EditorWorkflowEdgeType.TRUE:
                workflow_dag[edge.source].add_true_edge(edge.target)
            case EditorWorkflowEdgeType.FALSE:
                workflow_dag[edge.source].add_false_edge(edge.target)

    return Workflow(
        workflow_id=editor_workflow_graph.workflow_id,
        workflow_name=editor_workflow_graph.workflow_name,
        is_active=editor_workflow_graph.is_active,
        workflow_dag=WorkflowDAG(
            name=editor_workflow_graph.workflow_name,
            description=editor_workflow_graph.description,
            start=WorkflowStart(triggers=triggers),
            dag=workflow_dag,
        ),
    )


def _build_editor_schedule_trigger(
    trigger: WorkflowScheduleTrigger,
) -> EditorScheduleTrigger:
    """
    Build editor schedule trigger.

    Args:
        trigger: Workflow schedule trigger.

    Returns:
        Editor schedule trigger.
    """
    default_args = [
        (arg.name, serialize_json_with_reference(arg.value))
        for arg in trigger.default_args
    ]

    if trigger.cron:
        return EditorScheduleTrigger(
            schedule_type=EditorScheduleType.CRON,
            value=trigger.cron,
            default_args=default_args,
        )

    if trigger.interval_seconds:
        return EditorScheduleTrigger(
            schedule_type=EditorScheduleType.INTERVAL_SECONDS,
            value=str(trigger.interval_seconds),
            default_args=default_args,
        )

    if trigger.interval_minutes:
        return EditorScheduleTrigger(
            schedule_type=EditorScheduleType.INTERVAL_MINUTES,
            value=str(trigger.interval_minutes),
            default_args=default_args,
        )

    if trigger.interval_hours:
        return EditorScheduleTrigger(
            schedule_type=EditorScheduleType.INTERVAL_HOURS,
            value=str(trigger.interval_hours),
            default_args=default_args,
        )

    if trigger.interval_days:
        return EditorScheduleTrigger(
            schedule_type=EditorScheduleType.INTERVAL_DAYS,
            value=str(trigger.interval_days),
            default_args=default_args,
        )

    raise ValueError("Unhandled schedule trigger.")


def _build_workflow_schedule_trigger(
    trigger: EditorScheduleTrigger,
) -> WorkflowScheduleTrigger:
    """
    Build workflow schedule trigger.

    Args:
        trigger: Editor schedule trigger.

    Returns:
        Workflow schedule trigger.
    """
    default_args = [
        WorkflowDefaultArgument(name=k, value=deserialize_json_with_reference(v))
        for k, v in trigger.default_args
    ]

    match trigger.schedule_type:
        case EditorScheduleType.CRON:
            return WorkflowScheduleTrigger(
                cron=trigger.value, default_args=default_args
            )

        case EditorScheduleType.INTERVAL_SECONDS:
            return WorkflowScheduleTrigger(
                interval_seconds=int(trigger.value), default_args=default_args
            )

        case EditorScheduleType.INTERVAL_MINUTES:
            return WorkflowScheduleTrigger(
                interval_minutes=int(trigger.value), default_args=default_args
            )

        case EditorScheduleType.INTERVAL_HOURS:
            return WorkflowScheduleTrigger(
                interval_hours=int(trigger.value), default_args=default_args
            )

        case EditorScheduleType.INTERVAL_DAYS:
            return WorkflowScheduleTrigger(
                interval_days=int(trigger.value), default_args=default_args
            )

        case _:
            raise ValueError("Unhandled schedule type.")



from sqlmodel import SQLModel, Field
from sqlalchemy import TIMESTAMP
from sqlalchemy.sql.expression import func
from datetime import datetime
from abc import abstractmethod
from pydantic import BaseModel
from uuid import uuid4


class BaseSchema(SQLModel, table=False):
    created_at: datetime = Field(
        sa_type=TIMESTAMP(), sa_column_kwargs=dict(server_default=func.now())
    )
    updated_at: datetime = Field(
        sa_type=TIMESTAMP(), sa_column_kwargs=dict(server_default=func.now())
    )

    @abstractmethod
    def to_model(self, include_resources: bool = False) -> BaseModel:
        """
        TODO: add docstring
        """
        ...

    @staticmethod
    def generate_uuid4() -> str:
        return str(uuid4())



from fastapi import APIRouter, status, HTTPException
from collections import defaultdict
from sqlalchemy.exc import IntegrityError

from admyral.action_registry import ActionRegistry
from admyral.server.deps import get_admyral_store
from admyral.models import (
    ActionMetadata,
    ActionNamespace,
    EditorActions,
    EditorWorkflowGraph,
    WorkflowPushRequest,
    WorkflowPushResponse,
)
from admyral.editor import (
    workflow_to_editor_workflow_graph,
    editor_workflow_graph_to_workflow,
)
from admyral.server.endpoints.workflow_endpoints import push_workflow_impl


router = APIRouter()


@router.get("/actions", status_code=status.HTTP_200_OK)
async def load_workflow_actions() -> EditorActions:
    """
    Load all available workflow actions.
    """
    actions = [action.to_metadata() for action in ActionRegistry.get_actions()]
    custom_actions = await get_admyral_store().list_actions()
    actions.extend(custom_actions)

    actions_by_namespace = defaultdict(list)
    for action in actions:
        actions_by_namespace[action.display_namespace].append(action)

    # TODO: implement other control flow actions
    control_flow_actions = [
        # ActionMetadata(
        #     action_type="transform",
        #     display_name="Transform",
        #     display_namespace="Control Flow",
        # ),
        ActionMetadata(
            action_type="if_condition",
            display_name="If Condition",
            display_namespace="Control Flow",
        ),
        # ActionMetadata(
        #     action_type="python",
        #     display_name="Python",
        #     display_namespace="Control Flow",
        # ),
        # ActionMetadata(action_type="for_loop", display_name="For Loop"),
        # ActionMetadata(
        #     action_type="note",
        #     display_name="Note",
        #     display_namespace="Control Flow",
        # ),
    ]

    return EditorActions(
        control_flow_actions=control_flow_actions,
        namespaces=[
            ActionNamespace(namespace=namespace, actions=actions)
            for namespace, actions in actions_by_namespace.items()
        ],
    )


@router.get("/workflow", status_code=status.HTTP_200_OK)
async def load_workflow_as_react_flow_graph(workflow_id: str) -> EditorWorkflowGraph:
    """
    Load a workflow as a ReactFlow graph.
    """
    workflow = await get_admyral_store().get_workflow_by_id(workflow_id)
    if not workflow:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f'Workflow with the ID "{workflow_id}" does not exist.',
        )
    webhook = await get_admyral_store().get_webhook_for_workflow(workflow_id)
    return workflow_to_editor_workflow_graph(
        workflow,
        webhook_id=webhook.webhook_id if webhook else None,
        webhook_secret=webhook.webhook_secret if webhook else None,
    )


# TODO: TEST UNIQUENESS CONSTRAINT REGARDING WORKFLOW NAME
@router.post("/workflow", status_code=status.HTTP_201_CREATED)
async def save_workflow_from_react_flow_graph(
    editor_workflow_graph: EditorWorkflowGraph,
) -> WorkflowPushResponse:
    """
    Save a workflow from a ReactFlow graph.
    """
    workflow = editor_workflow_graph_to_workflow(editor_workflow_graph)
    try:
        return await push_workflow_impl(
            workflow_name=workflow.workflow_name,
            workflow_id=workflow.workflow_id,
            request=WorkflowPushRequest(
                workflow_dag=workflow.workflow_dag, activate=workflow.is_active
            ),
        )
    except IntegrityError:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"A workflow with the name '{workflow.workflow_name}' already exists. Workflow names must be unique.",
        )



from logging.config import fileConfig

import asyncio
import os
from sqlalchemy.engine import Connection
from sqlalchemy.ext.asyncio import async_engine_from_config
from sqlalchemy import pool
from sqlmodel import SQLModel
from dotenv import load_dotenv
import aiofiles.os

from alembic import context

from admyral.db.schemas import *  # noqa F403

load_dotenv()

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
# target_metadata = None
target_metadata = SQLModel.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url", os.environ["MIGRATION_DATABASE_URL"])

    # if we have a SQLite database, we need to make sure that its parent directory exists.
    if url.startswith("sqlite"):
        db_parent_path = os.path.dirname(url.split("///")[1])
        path_exists = os.path.exists(db_parent_path)
        if not path_exists:
            os.makedirs(db_parent_path, exist_ok=True)

    context.configure(
        url=url,
        target_metadata=target_metadata,
        version_table_schema=target_metadata.schema,
        literal_binds=True,
        include_schemas=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection: Connection) -> None:
    cfg = config.get_section(config.config_ini_section)
    cfg["sqlalchemy.url"] = cfg.get(
        "sqlalchemy.url", os.environ.get("MIGRATION_DATABASE_URL")
    )

    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        version_table_schema=target_metadata.schema,
        include_schemas=True,
    )

    with context.begin_transaction():
        context.run_migrations()


async def run_async_migrations() -> None:
    """In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    cfg = config.get_section(config.config_ini_section, {})
    cfg["sqlalchemy.url"] = cfg.get(
        "sqlalchemy.url", os.environ.get("MIGRATION_DATABASE_URL")
    )

    # if we have a SQLite database, we need to make sure that its parent directory exists.
    db_url = cfg["sqlalchemy.url"]
    if db_url.startswith("sqlite"):
        db_parent_path = os.path.dirname(db_url.split("///")[1])
        path_exists = await aiofiles.os.path.exists(db_parent_path)
        if not path_exists:
            await aiofiles.os.makedirs(db_parent_path, exist_ok=True)

    connectable = async_engine_from_config(
        cfg,
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    async with connectable.connect() as connection:
        await connection.run_sync(do_run_migrations)

    await connectable.dispose()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode."""

    asyncio.run(run_async_migrations())


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import asyncio
import argparse

from admyral.services.api_service import run_api
from admyral.services.worker_service import launch_worker


def parse_args() -> dict:
    parser = argparse.ArgumentParser(description="Admyral")
    parser.add_argument(
        "exec_type",
        choices=["api", "worker"],
    )
    parser.add_argument(
        "--docker",
        default=False,
        action="store_true",
    )
    return dict(parser.parse_args()._get_kwargs())


def worker(args: dict) -> None:
    asyncio.run(launch_worker(args))


def api(args: dict) -> None:
    asyncio.run(run_api(args))


def main() -> None:
    args = parse_args()
    match args["exec_type"]:
        case "worker":
            worker(args)
        case "api":
            api(args)


if __name__ == "__main__":
    main()



from sqlmodel import Field, Relationship
from sqlalchemy import TEXT, ForeignKeyConstraint
from typing import TYPE_CHECKING

from admyral.db.schemas.base_schemas import BaseSchema
from admyral.models import WorkflowWebhook

if TYPE_CHECKING:
    from admyral.db.schemas.workflow_schemas import WorkflowSchema


class WorkflowWebhookSchema(BaseSchema, table=True):
    """
    Schema for Workflow Webhook
    """

    __tablename__ = "workflow_webhooks"

    __table_args__ = (
        ForeignKeyConstraint(
            ["workflow_id"],
            ["workflows.workflow_id"],
            ondelete="CASCADE",
        ),
    )

    # primary keys
    webhook_id: str = Field(
        sa_type=TEXT(), primary_key=True, default_factory=BaseSchema.generate_uuid4
    )

    # foreign keys
    workflow_id: str = Field(sa_type=TEXT(), index=True)

    # other fields
    webhook_secret: str = Field(sa_type=TEXT())

    # relationship parents
    workflow: "WorkflowSchema" = Relationship(back_populates="webhooks")

    def to_model(self, include_resources: bool = False) -> WorkflowWebhook:
        return WorkflowWebhook.model_validate(
            {
                "webhook_id": self.webhook_id,
                "workflow_id": self.workflow_id,
                "webhook_secret": self.webhook_secret,
            }
        )



from fastapi import APIRouter, status, HTTPException

from admyral.models import WorkflowRunMetadata, WorkflowRunStepMetadata, WorkflowRunStep
from admyral.server.deps import get_admyral_store


router = APIRouter()


@router.get("/{workflow_id}", status_code=status.HTTP_200_OK)
async def list_workflow_runs(workflow_id: str) -> list[WorkflowRunMetadata]:
    """
    List all workflow runs.

    Args:
        workflow_id: The workflow id

    Returns:
        A list of workflow runs.
    """
    return await get_admyral_store().list_workflow_runs(workflow_id)


@router.get("/{workflow_id}/{workflow_run_id}", status_code=status.HTTP_200_OK)
async def get_workflow_run_steps(
    workflow_id: str, workflow_run_id: str
) -> list[WorkflowRunStepMetadata]:
    """
    Get workflow runs by workflow id.

    Args:
        workflow_id: The workflow id
        workflow_run_id: The workflow run id

    Returns:
        A list of workflow run steps.
    """
    return await get_admyral_store().list_workflow_run_steps(
        workflow_id, workflow_run_id
    )


@router.get(
    "/{workflow_id}/{workflow_run_id}/{workflow_run_step_id}",
    status_code=status.HTTP_200_OK,
)
async def get_workflow_run_step(
    workflow_id: str, workflow_run_id: str, workflow_run_step_id: str
) -> WorkflowRunStep:
    """
    Get workflow run step.

    Args:
        workflow_id: The workflow id
        workflow_run_id: The workflow run id
        workflow_run_step_id: The workflow run step id

    Returns:
        The workflow run step.
    """
    step = await get_admyral_store().get_workflow_run_step(
        workflow_id, workflow_run_id, workflow_run_step_id
    )
    if not step:
        raise HTTPException(status_code=404, detail="Step not found")
    return step



from typing import Annotated
from httpx import Client

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue


def get_alienvault_otx_client(api_key: str) -> Client:
    return Client(
        base_url="https://otx.alienvault.com/api/v1",
        headers={
            "X-OTX-API-KEY": api_key,
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Analyze Domain",
    display_namespace="AlienVault OTX",
    description="Analyze a domain using AlienVault OTX",
    secrets_placeholders=["ALIENVAULT_OTX_SECRET"],
)
def alienvault_otx_analyze_domain(
    domain: Annotated[
        str,
        ArgumentMetadata(
            display_name="Domain",
            description="The domain to analyze",
        ),
    ],
) -> JsonValue:
    secret = ctx.get().secrets.get("ALIENVAULT_OTX_SECRET")
    api_key = secret["api_key"]

    with get_alienvault_otx_client(api_key) as client:
        response = client.get(
            f"/indicators/domain/{domain}/general",
            headers={
                "X-OTX-API-KEY": api_key,
                "Content-Type": "application/json",
            },
        )
        response.raise_for_status()
        return response.json()



import asyncio
from typing import Any


# Global variable to store the main event loop reference
MAIN_EVENT_LOOP = None


def capture_main_event_loop():
    global MAIN_EVENT_LOOP
    MAIN_EVENT_LOOP = asyncio.get_event_loop()


def execute_future(func: asyncio.coroutines) -> Any:
    assert MAIN_EVENT_LOOP is not None, "Main event loop not captured."
    future = asyncio.run_coroutine_threadsafe(func, MAIN_EVENT_LOOP)
    return future.result()



from datetime import datetime
from sqlmodel import Field
from sqlalchemy import TEXT, TIMESTAMP

from admyral.db.schemas.base_schemas import BaseSchema
from admyral.models import PipLockfile


class PipLockfileCacheSchema(BaseSchema, table=True):
    """
    Schema for pip lockfile cache
    """

    __tablename__ = "pip_lockfile_cache"

    # primary keys
    hash: str = Field(sa_type=TEXT(), primary_key=True)

    # other fields
    lockfile: str = Field(sa_type=TEXT())
    expiration_time: datetime = Field(sa_type=TIMESTAMP())

    def to_model(self) -> PipLockfile:
        return PipLockfile.model_validate(
            {
                "hash": self.hash,
                "lockfile": self.lockfile,
                "expiration_time": self.expiration_time_unix,
            }
        )

    @property
    def expiration_time_unix(self) -> int:
        return int(self.expiration_time.timestamp())

    @expiration_time_unix.setter
    def expiration_time_unix(self, value: int):
        self.expiration_time = datetime.fromtimestamp(value)



from abc import abstractmethod
from typing import Optional
import json

from admyral.db.store_interface import StoreInterface
from admyral.utils.crypto import decrypt_secret, encrypt_secret
from admyral.models import Secret, SecretMetadata
from admyral.config.config import GlobalConfig, SecretsManagerType


class SecretsManager:
    @abstractmethod
    async def get(self, user_id: str, secret_id: str) -> Optional[Secret]:
        """
        Retrieve a secret for a user. If the secret does not exist, return None.

        Args:
            user_id: The user id of the user whose secret is retrieved.
            secret_id: The id of the secret to retrieve.
        """
        raise NotImplementedError("get method not implemented")

    @abstractmethod
    async def set(self, user_id: str, secret: Secret) -> None:
        """
        Set a secret for a user. If the secret already exists, it is overwritten.
        The function takes care of encrypting the secret.

        Args:
            user_id: The user id of the user whose secret is set.
            secret: The secret to set.
        """
        raise NotImplementedError("set method not implemented")

    @abstractmethod
    async def delete(self, user_id: str, secret_id: str) -> None:
        """
        Delete a secret for a user.

        Args:
            user_id: The user id of the user whose secret is deleted.
            secret_id: The id of the secret to delete.
        """
        raise NotImplementedError("delete method not implemented")

    @abstractmethod
    async def list(self, user_id: str) -> list[SecretMetadata]:
        """
        List the secret names for a user.

        Args:
            user_id: The user id of the user whose secrets are listed.
        """
        raise NotImplementedError("list method not implemented")


class SQLSecretsManager(SecretsManager):
    def __init__(self, db: StoreInterface) -> None:
        self.db = db

    async def get(self, user_id: str, secret_id: str) -> Secret:
        encrypted_secret = await self.db.get_secret(user_id, secret_id)
        return (
            Secret(
                secret_id=secret_id,
                secret=json.loads(decrypt_secret(encrypted_secret.encrypted_secret)),
            )
            if encrypted_secret
            else None
        )

    async def set(self, user_id: str, secret: Secret) -> None:
        serialized_secret = json.dumps(secret.secret)
        encrypted_secret = encrypt_secret(serialized_secret)
        secret_schema = list(secret.secret.keys())
        return await self.db.store_secret(
            user_id, secret.secret_id, encrypted_secret, secret_schema
        )

    async def delete(self, user_id: str, secret_id: str) -> None:
        return await self.db.delete_secret(user_id, secret_id)

    async def list(self, user_id: str) -> list[SecretMetadata]:
        return await self.db.list_secrets(user_id)


def secrets_manager_factory(db: StoreInterface) -> SecretsManager:
    secrets_manager_type = GlobalConfig().secrets_manager_type
    match secrets_manager_type:
        case SecretsManagerType.SQL:
            return SQLSecretsManager(db)

        case _:
            raise ValueError(f"Unknown secrets manager type: {secrets_manager_type}")



import asyncio
from temporalio.client import Client
from typing import Coroutine

from admyral.logger import get_logger

logger = get_logger(__name__)


async def wait_for_temporal(
    target_host: str, max_retries: int = 10, sleep_time_in_sec: float = 0.5
) -> None:
    """
    Wait for Temporal to start up and be ready to accept connections.

    Args:
        target_host: The host and port of the Temporal server.
        max_retries: The maximum number of retries before giving up.
        sleep_time_in_sec: The time to sleep between retries.
    """
    for _ in range(max_retries):
        await asyncio.sleep(sleep_time_in_sec)
        try:
            client = await Client.connect(target_host)
            break
        except Exception:
            client = None

    if client is None or not await client.service_client.check_health():
        raise Exception("Failed to start Temporal!")

    logger.info("Temporal started successfully.")


async def start_after_temporal(
    tasks: list[Coroutine],
    target_host: str,
    max_retries: int = 10,
    sleep_time_in_sec: float = 0.5,
) -> None:
    """
    Start the backend server and the worker after Temporal has started.

    Args:
        target_host: The host and port of the Temporal server.
    """
    await wait_for_temporal(
        target_host, max_retries=max_retries, sleep_time_in_sec=sleep_time_in_sec
    )
    logger.info("Temporal started. Starting dependent tasks...")
    await asyncio.gather(*tasks)


async def run_temporal():
    """
    Start a local Temporal instance.
    """
    proc = await asyncio.create_subprocess_exec(
        "temporal",
        "server",
        "start-dev",
        "--headless",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    logger.info("Temporal started.")
    await proc.communicate()
